import datetime

from enum import Enum
from threading import Condition
from queue import Queue
from TDhelper.Event.Event import Event
from TDhelper.Spider.models.Cache_L2 import L2, RecordStatus

thread_condition = Condition()


class L1(Event):
    def __init__(self, size=10):
        super(L1, self).__init__()
        self._L1_Size = size
        self._L1 = Queue(size)
        self._L2= L2()

    def push(self, value, key, source_url=None):
        try:
            thread_condition.acquire()
            if not self._L1.full():
                self._L1.put([value, source_url, key])
                self.on(event.onPush, self)
            else:
                # L1缓存数据存满了，将数据缓存到L2
                self._L2.URL = value
                self._L2.Source = source_url
                self._L2.status = RecordStatus.WAIT.value
                self._L2.Key= key
                self._L2.updatetime= datetime.datetime.now()
                self._L2.toSave()
                self._L2.clear()
        finally:
            thread_condition.release()

    def pop(self):
        try:
            thread_condition.acquire()
            if self._L1.empty():
                # L1缓存数据读取完毕，从L2缓存读取数据
                if not self._getL2cache(size= self._L1_Size):
                    return None
            return self._L1.get()
        finally:
            thread_condition.release()

    def _getL2cache(self, size=1, type= 0):
         # L1缓存数据读取完毕，从L2缓存读取数据
        m_records = self._L2.find(**{'limit': size})
        if m_records.count() > 0:
            for item in m_records:
                if type==0:
                    self.push(item['url'], item['key'], item['source'])
                else:
                    self._L1.put([item['url'], item['source'], item['key']])
                self._L2.model = item
                self._L2.deleteById()
                self._L2.clear()
            return True
        else:
            return False

    def getCache(self):
        '''
            不会触发push事件，仅返回cache本身.取出后会从缓存删除
        '''
        try:
            thread_condition.acquire()
            if self._L1.empty():
                # L1缓存数据读取完毕，从L2缓存读取数据
                if not self._getL2cache(type= 1):
                    return None
            return self._L1.get()
        finally:
            thread_condition.release()


    def getSize(self):
        return self.L1.qsize()


class event(Enum):
    onPush = 'onPush'
    onPop = 'onPop'



from .meta.modelMeta import modelMeta

class model(metaclass=modelMeta):
    __excelHandle__ = None  # excel object.
    __sheetHandle__ = None  # sheet object.
    __rows__ = []  # all rows.

    def __init__(self, excelPath, headrow=True):
        assert excelPath, "parameter <excelPath> is None."
        self.Meta.file = excelPath
        self.Meta.headrow = headrow
        try:
            self.__initExcelHandle__()
        except Exception as e:
            raise e

    def __initExcelHandle__(self):
        return None

    class Meta:
        """
        元数据
        - file: <string>, 文件路径.
        - sheet: <string>, sheet名称.
        - extension: <string>, 文件类型(xlsx,csv)
        """

        file = ""
        sheet = "sheet1"
        extension = "xlsx"
        headrow = True



from TDhelper.Event.classEvent.meta import eventMeta
def hook(before= None, complete= None):
    '''
    函数钩子

    - 参数
        - before: function, 执行前需要执行的函数. 如果函数返回值为*args,**kwargs则传如参数会对应改变.
        - complete: function, 执行完成需要执行的函数. 可更改返回结果. 
    '''
    def wrapper(func):
        def deco(*args, **kwargs):
            try:
                m_args= None
                if before:
                    m_args, kwargs= before(*args, **kwargs)
                if m_args:
                    ret= func(*m_args, **kwargs)
                else:
                    ret= func(*args, **kwargs)
                if complete:
                    kwargs['func-result']= ret
                    if m_args:
                        ret= complete(*m_args, **kwargs)
                    else:
                        ret= complete(*args, **kwargs)
                return ret
            except Exception as e:
                raise e
        return deco
    return wrapper

class event(metaclass= eventMeta):

    def _event_register(self, name, func):
        pass

    def _event_remove(self, name):
        pass

    def _event_on(self, name, *args, **kwargs):
        pass
    


from TDhelper.generic.transformationType import transformation
from TDhelper.MagicCls.FieldsType import FieldType

class _AttributeOverride:
    def __init__(self, name, m_type):
        self._name = name
        self._type = m_type

    def __get__(self, instance, owen):
        return instance.__dict__[self._name]

    def __set__(self, instance, value):
        instance.__dict__[self._name] = transformation(value, self._type)

    def __delete__(self, instance):
        instance.__dict__.pop(self._name)


class MagicMeta(type):
    def __new__(cls, name, bases, dct):
        attrs = {
            "__mapping__": {},
            "__reverse_mapping__": {},
            "__sourceOffset__": 0,
            "__instance__": __instance__,
            "__items__": [],
            "items": items,
        }
        for name, value in dct.items():
            if isinstance(dct[name], FieldType):
                try:
                    assert not attrs.__contains__(name),"field name<%s> is default attrbute, please change." % name
                    attrs["__mapping__"][name] = value.bind
                    attrs["__reverse_mapping__"][value.bind] = name
                    attrs[name] = _AttributeOverride(name, value.fieldType)
                except Exception as e:
                    raise e
            else:
                attrs[name] = value
        return super(MagicMeta, cls).__new__(cls, name, bases, attrs)


def __instance__(self):
    for n, v in self.__mapping__.items():
        setattr(self, n, self.Meta.source[self.__sourceOffset__][v])
    return self

def items(self):
    for r in self.Meta.source:
        for n, v in r.items():
            setattr(self, self.__reverse_mapping__[n], v)
        yield self



import os

class stream:
    '''
    未完成
    '''
    def __init__(self, streamPath):
        self._handle= None
        if self._exsiteFolder(streamPath):
            self._handle= os.open(streamPath)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._handle.close()

    def _exsiteFolder(self,streamPath):
        if not os.path.exists(streamPath.rsplit(r'\\',1)[0]):
            try:
                os.mkdir(streamPath)
                return True
            except Exception as e:
                return False
        else:
            return True